@@ -24,7 +24,6 @@ import time
 from typing import (Callable, Optional, Sequence, Type, Union)  # noqa: H301
 import uuid as uuid_lib
 
-
 from oslo_concurrency import processutils as putils
 from oslo_log import log as logging
 
@@ -134,6 +133,9 @@ class Portal(object):
     """Representation of an NVMe-oF Portal with some related operations."""
     LIVE = 'live'
     MISSING = None  # Unkown or not present in the system
+    CONNECTING = 'connecting'
+    # Default value of reconnect_delay in sysfs
+    DEFAULT_RECONNECT_DELAY = 10
     controller: Optional[str] = None  # Don't know controller name on start
 
     def __str__(self) -> str:
@@ -177,6 +179,15 @@ class Portal(object):
             return ctrl_property('state', self.controller)
         return None
 
+    @property
+    def reconnect_delay(self) -> int:
+        # 10 seconds is the default value of reconnect_delay
+        if self.controller:
+            res = ctrl_property('reconnect_delay', self.controller)
+            if res is not None:
+                return int(res)
+        return self.DEFAULT_RECONNECT_DELAY
+
     def get_device(self) -> Optional[str]:
         """Get a device path using available volume identification markers.
 
@@ -387,8 +398,9 @@ class Target(object):
         # Unlike "nvme list-subsys -o json" sysfs addr is separated by a comma
         sysfs_portals: list[tuple[Optional[str],
                                   Optional[str],
+                                  Optional[Union[str, utils.Anything]],
                                   Optional[Union[str, utils.Anything]]]] = [
-            (f'traddr={p.address},trsvcid={p.port}', p.transport, hostnqn)
+            (p.address, p.port, p.transport, hostnqn)
             for p in self.portals
         ]
         known_names: list[str] = [p.controller for p in self.portals
@@ -410,7 +422,19 @@ class Target(object):
 
             # The right subsystem, but must also be the right portal
             ctrl_transport = sysfs_property('transport', ctrl_path)
-            ctrl_addr = sysfs_property('address', ctrl_path)
+
+            # Address in sysfs may contain src_addr in some systems. Parse and
+            # only use destination addr and port
+            address = sysfs_property('address', ctrl_path)
+            if not address:
+                LOG.error("Couldn't read address for %s", ctrl_path)
+                continue
+            ctrl_address = dict((x.split('=')
+                                 for x in address.split(',')))
+
+            ctrl_addr = ctrl_address['traddr']
+            ctrl_port = ctrl_address['trsvcid']
+
             # hostnqn value not present in all OSs.  Ignore when not present
             ctrl_hostnqn = sysfs_property('hostnqn', ctrl_path) or utils.ANY
             # Warn once per target for OS not presenting the hostnqn on sysfs
@@ -419,7 +443,7 @@ class Target(object):
                             "Controller may be incorrectly matched.")
                 warned = True
 
-            ctrl_portal = (ctrl_addr, ctrl_transport, ctrl_hostnqn)
+            ctrl_portal = (ctrl_addr, ctrl_port, ctrl_transport, ctrl_hostnqn)
             try:
                 index = sysfs_portals.index(ctrl_portal)
 
@@ -684,6 +708,8 @@ class NVMeOFConnector(base.BaseLinuxConnector):
 
     # Use a class attribute since a restart is needed to change it on the host
     native_multipath_supported = None
+    # Time we think is more than reasonable to establish an NVMe-oF connection
+    TIME_TO_CONNECT = 10
 
     def __init__(self,
                  root_helper: str,
@@ -768,10 +794,16 @@ class NVMeOFConnector(base.BaseLinuxConnector):
         ret = {}
 
         nqn = None
+        hostid = None
         uuid = nvmf._get_host_uuid()
         suuid = priv_nvmeof.get_system_uuid()
         if cls.nvme_present():
-            nqn = utils.get_host_nqn()
+            nqn = utils.get_host_nqn(suuid)
+            # Ensure /etc/nvme/hostid exists and defaults to the system uuid,
+            # or a random value.
+            hostid = utils.get_nvme_host_id(suuid)
+        if hostid:
+            ret['nvme_hostid'] = hostid
         if uuid:
             ret['uuid'] = uuid
         if suuid:
@@ -783,12 +815,17 @@ class NVMeOFConnector(base.BaseLinuxConnector):
 
     def _get_host_uuid(self) -> Optional[str]:
         """Get the UUID of the first mounted filesystem."""
-        cmd = ('findmnt', '/', '-n', '-o', 'SOURCE')
+        cmd = ('findmnt', '-v', '/', '-n', '-o', 'SOURCE')
         try:
             lines, err = self._execute(
                 *cmd, run_as_root=True, root_helper=self._root_helper)
+            source = lines.split('\n')[0]
+            # In a container this could be 'overlay', which causes the blkid
+            # command to fail.
+            if source == "overlay":
+                return None
             blkid_cmd = (
-                'blkid', lines.split('\n')[0], '-s', 'UUID', '-o', 'value')
+                'blkid', source, '-s', 'UUID', '-o', 'value')
             lines, _err = self._execute(
                 *blkid_cmd, run_as_root=True, root_helper=self._root_helper)
             return lines.split('\n')[0]
@@ -856,6 +893,7 @@ class NVMeOFConnector(base.BaseLinuxConnector):
         """
         connected = False
         missing_portals = []
+        reconnecting_portals = []
 
         for portal in target.portals:
             state = portal.state  # store it so we read only once from sysfs
@@ -866,6 +904,9 @@ class NVMeOFConnector(base.BaseLinuxConnector):
             # Remember portals that are not present in the system
             elif state == portal.MISSING:
                 missing_portals.append(portal)
+            elif state == portal.CONNECTING:
+                LOG.debug('%s is reconnecting', portal)
+                reconnecting_portals.append(portal)
             # Ignore reconnecting/dead portals
             else:
                 LOG.debug('%s exists but is %s', portal, state)
@@ -890,8 +931,12 @@ class NVMeOFConnector(base.BaseLinuxConnector):
                     # c-vol running on same node with different lock paths) or
                     # an admin is touching things manually. Not passing these
                     # exit codes in check_exit_code parameter to _execute so we
-                    # can log it.
-                    if exc.exit_code not in (70, errno.EALREADY):
+                    # can log it.  nvme cli v2 returns 1, so we parse the
+                    # message. Some nvme cli versions return errors in stdout,
+                    # so we look in stderr and stdout.
+                    if not (exc.exit_code in (70, errno.EALREADY) or
+                            (exc.exit_code == 1 and
+                             'already connected' in exc.stderr + exc.stdout)):
                         LOG.error('Could not connect to %s: exit_code: %s, '
                                   'stdout: "%s", stderr: "%s",', portal,
                                   exc.exit_code, exc.stdout, exc.stderr)
@@ -900,11 +945,32 @@ class NVMeOFConnector(base.BaseLinuxConnector):
                     LOG.warning('Race condition with some other application '
                                 'when connecting to %s, please check your '
                                 'system configuration.', portal)
-                    connected = True
+                    state = portal.state
+                    if state == portal.LIVE:
+                        connected = True
+                    elif state == portal.CONNECTING:
+                        reconnecting_portals.append(portal)
+                    else:
+                        LOG.error('Ignoring %s due to unknown state (%s)',
+                                  portal, state)
 
                 if not do_multipath:
                     break  # We are connected
 
+        if not connected and reconnecting_portals:
+            delay = self.TIME_TO_CONNECT + max(p.reconnect_delay
+                                               for p in reconnecting_portals)
+            LOG.debug('Waiting %s seconds for some nvme controllers to '
+                      'reconnect', delay)
+            timeout = time.time() + delay
+            while time.time() < timeout:
+                time.sleep(1)
+                if any(p.is_live for p in reconnecting_portals):
+                    LOG.debug('Reconnected')
+                    connected = True
+                    break
+            LOG.debug('No controller reconnected')
+
         if not connected:
             raise exception.VolumeDeviceNotFound(device=target.nqn)
 
@@ -1195,7 +1261,11 @@ class NVMeOFConnector(base.BaseLinuxConnector):
             LOG.info('Wait 5 seconds and return whatever size is present')
             time.sleep(5)
 
-        return utils.get_device_size(self, device_path)
+        size = utils.get_device_size(self, device_path)
+        if size is None:
+            raise exception.BrickException(
+                'get_device_size returned non-numeric size')
+        return size
 
     # #######  RAID methods ########
 
